#include <wgd/game.h>
#include <wgd/scene.h>
#include <wgd/keyboard.h>
#include <wgd/camera.h>

#include "player.h"

using namespace wgd;

Player::Player(const doste::OID &id) 
 : IModel(id), prejump(false), canMove(true), direction(1), plat(NULL) {
	DOSTE_REGISTER(100, getID()("jumpHeight"));
	DOSTE_REGISTER(101, getID()("gravity"));
	
	setup();
}

void Player::update(){
	float t = Game::frameTime();

	//yaw
	angle(Vector3D(0, (plane()+2) * M_PI_2, 0));

	//jump
	if(standing && canMove && Keyboard::keyPressed(Keyboard::KEY_UP)) jump();
	if(prejump && frame()>0.8){
		vel.y = jumpVelocity;
		
		//inherit velocity from platform
		if(standing){
			Vector2D iVel = plat->velocityAt(pos());
			if(plane()<2) iVel.x = -iVel.x;
			vel = vel + iVel;
		}
		
		prejump=false;
		canMove=true;
		standing=false;
	}

	//movement
	float movement=0;
	if(Keyboard::keyPressed(Keyboard::KEY_LEFT)) {
		goLeft();
		if(canMove) {
			movement = 50 * -t;
			if(vel.x > 0) vel.x=0;
		}
	} else if(Keyboard::keyPressed(Keyboard::KEY_RIGHT)) {
		goRight();
		if(canMove){
			movement = 50 * t;
			if(vel.x <0) vel.x=0;
		}
	} else {
		stop();
		if(standing) vel.x=0;
	}
	
	//inherited velcity
	//if(standing) vel = plat->velocityAt(pos());
	
	//rotation from gear
	if(standing) pos(plat->playerPos(pAngle, pRadius));
	
	
	//apply gravity and get movement vector (for framerate independance)
	Vector2D move;
	vel.y = vel.y - gravity() * t;
	
	move.x = (vel.x+movement) * t;
	move.y = vel.y * t - gravity() * t * t / 2;
	
	//reverse controls
	if(plane()<2) move.x = -move.x;
	
	
	//3d position
	Vector3D nPos3D = position();
	nPos3D.y += move.y;
	switch(plane()){
		case 0: nPos3D.x -= move.x; break;
		case 1: nPos3D.z -= move.x; break;
		case 2: nPos3D.x += move.x; break;
		case 3: nPos3D.z += move.x; break;
	}
		
	
	//if more than one colision then maybe get squished
	int collisions=0;
	
	//Collision checks
	Collision c[10];
	int num = Scene::current()->colliding(this, nPos3D, c, 10);
	for(int i=0; i<num; i++){
		if(c[i].instance->get("type")=="platform"){
			if(usePlatform((Gear*)c[i].instance, move)) collisions++;
		}
	}
	
	//what to do here??
	if(collisions>1) cout << "Multiple collisions\n";
	
	//straight falling
	if(collisions==0){
		if(!standing) pos(pos()+move);
		//standing = false;
	}
	
	//Die if fallen off the bottom of the screen
	if(pos().y < Camera::camera()->position().y - 5) die();
	
	//animate model
	IModel::update();
}

bool Player::usePlatform(Gear* gear, Vector2D &move){
	
	//new position
	Vector2D nPos = pos() + move*3;
	Vector2D back = pos() - move*6;	
	
	gear->test = Vector3D(0,0,0);
	
	//if side on, all the values will be different
	if(gear->plane() != plane()){
		Vector2D dPos = pos();
		//get new x coordinate
		if(gear->plane()==1 || gear->plane()==3) dPos.x = position().z;
		else dPos.x = position().x;
		
		nPos = dPos + move*3;
		back = dPos - move*6;
		
	}
	
	Vector2D tangent;
	bool hit = gear->intersect(back, nPos, tangent);
	if(hit || (standing && plat==gear)){
		
		if(!hit){
			
			//if(gear->inside(pos())) cout << "STUCK\n";
			//if(gear->inside(back)) cout << "BAD POSITION\n";
			//else{
				
				//must get the tangent
				nPos = pos(); nPos.y-= 0.1;
				Vector3D back2 = pos(); back2.y+=0.1;
				if(!gear->intersect(back2, nPos, tangent)){
					standing=false;
					return false;
				}
			//}
			
		}
		
		//draw velocity
		gear->velocityAt(pos());
		
		tangent = tangent.normalise();
		
		//slide?
		Vector2D tn(tangent.y, -tangent.x);
		//if(tn.dot(move.normalise()) > -0.7){
			float dot = tangent.dot(move.normalise());
			
			//get component vector [of the remainder]
			float d = dot * move.length();  //(pos+move-nPos).length();
			nPos = nPos + tangent * d;
			
			//check this motion
			gear->intersect(back, nPos, tangent);
		//}
		
		//set plane
		plane(gear->plane());

		//set player position
		pos(nPos);
		
		
		//standing
		vel.y=0;
		standing=true;
		
		//remember where you are standing
		Vector2D ctr = gear->centre2D();
		pAngle = atan2(nPos.y-ctr.y, nPos.x-ctr.x) + M_PI_2 - gear->rotation();
		pRadius = (nPos - ctr).length();
		plat = gear;

		return true;
		
	} else {
		return false;
	}
}

Vector2D Player::pos(){
	Vector3D p = position();
	switch(plane()){
		case 0: return Vector2D(p.x, p.y);
		case 1: return Vector2D(p.z, p.y);
		case 2: return Vector2D(p.x, p.y);
		case 3: return Vector2D(p.z, p.y);
	}
}

void Player::pos(const Vector2D &p){
	Vector3D v;
	if(standing) v = plat->position();
	
	switch(plane()){
		case 0: set(ix::x, p.x); if(standing) set(ix::z, v.z); break;
		case 1: set(ix::z, p.x); if(standing) set(ix::x, v.x); break;
		case 2: set(ix::x, p.x);  if(standing) set(ix::z, v.z); break;
		case 3: set(ix::z, p.x);  if(standing) set(ix::x, v.x); break;
	}
	set(ix::y, p.y);
}

void Player::die(){ 
	position(Vector3D(0, 1.5, 0)); 
	standing=false; 
	vel = Vector2D(0,0); 
}

void Player::jump(){ 
	if(direction== 1) animation("rjump");
	if(direction==-1) animation("ljump");
	prejump=true;
	canMove=false;
}
void Player::stop(){
	if(standing && canMove){
		if(direction==1) animation("ridle");
		if(direction==-1) animation("lidle");
	}
}
void Player::goLeft(){
	if(!standing){
		direction=-1;
	} else if(direction==1 && canMove){
		animation("turnleft");
		direction=-1;
		canMove=false;
	} else if(canMove || animation()=="lidle") {
		canMove=true;
		if(animation()!="lrun") animation("lstart");
		else animation("lrun");
	}
}
void Player::goRight(){
	if(!standing){
		direction=1;
	} else if(direction==-1 && canMove){
		animation("turnright");
		direction=1;
		canMove=false;
	} else if(canMove || animation()=="ridle") {
		canMove=true;
		if(animation()!="rrun") animation("rstart");
		else animation("rrun");
	}
}

void Player::fall(){
	if(vel.y <= 0) {
		if(direction==1) animation("rfall");
		if(direction==-1) animation("lfall");
	} else if(frame()>3) {
		if(direction==1) animation("rjump");
		if(direction==-1) animation("ljump");
		if(frame()<3) frame(3);
	}
}

void Player::land(){}


void Player::setJumpVelocity(DOSTE_HANDLER){
	jumpVelocity = sqrt(2 * jumpHeight() * gravity());
}
